/*

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

public class Block
{
	Block(Blockmap[] blocktable)
	{
		setPosition(0, 0);
		m_blocktable = blocktable;
	}
	
	Block(int x, int y, Blockmap[] blocktable)
	{
		setPosition(x, y);
		m_blocktable = blocktable;
	}
	
	public void translate(int dx, int dy)
	{
		m_x += dx;
		m_y += dy;
	}
	
	public void rotate(int dir)
	{
		setAngle(m_blocktableindex + dir);
	}
	
	public void setAngle(int angle)
	{
		if(angle >= m_blocktable.length)
			angle = 0;
		if(angle < 0)
			angle = m_blocktable.length-1;
			
		m_blocktableindex = angle;
	}
	
	public int getAngle()
	{
		return m_blocktableindex;	
	}	
		
	public int getX()
	{
		return m_x;
	}
	
	public int getY()
	{
		return m_y;
	}
	
	public void setValue(int value)
	{
		for(int i=0;i<m_blocktable.length;i++)
			m_blocktable[i].setValue(value);
	}
	
	public void invertValue()
	{
		for(int i=0;i<m_blocktable.length;i++)
			m_blocktable[i].invertValue();
	}
	
	public void setPosition(int x, int y)
	{
		m_x = x;
		m_y = y;
	}
	
	public Blockmap getMap()
	{
		return getMap(m_blocktableindex);
	}
	
	public Blockmap getMap(int index)
	{
		if(index < 0)
			index = 0;
		if(index >= m_blocktable.length)
			index = m_blocktable.length - 1;
			
		return m_blocktable[index];
	}
	
	public Block duplicate()
	{
		Blockmap[] blocktable = new Blockmap[m_blocktable.length];
		
		for(int i=0;i<m_blocktable.length;i++)
			blocktable[i] = m_blocktable[i].duplicate();
				
		Block bdup = new Block(getX(), getY(), blocktable);
		bdup.setAngle(getAngle());
		
		return bdup;
	}
	
	protected int m_x = 0, m_y = 0;
	protected Blockmap[] m_blocktable;
	protected int m_blocktableindex = 0;
}