/*

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

import java.awt.*;
import java.util.Random;

public class Tetris extends Canvas
{
	Tetris()
	{
		init(200, 400, 25);
	}

	Tetris(int width, int height, int blocksize)
	{
		init(width, height, blocksize);
	}

	// init

	public synchronized void init(int width, int height, int blocksize)
	{
		m_pause = true;
		m_pauseinterval = 0;
		m_width = width;
		m_height = height;
		m_blocksize = blocksize;
		
		m_map = new int[m_width][m_height];
		m_backbuffer = null;
		
		initBlocks();	// initBlocks steht am fileende (wegen der laenge)
		setBackground(getColor(0));
		setDrawScore(true);
		setDrawLinecount(true);
		setDrawLevel(true);
	}
	
	// public

	public synchronized void start()
	{
		setScore(0);
		setLinecount(0);
		setLevel(1);
		m_pause = false;
		clear();

		m_block = null;
		m_nextblock = null;
		createNewBlock();
		setSpeed(500);
	}
	public synchronized void stop()
	{
		stop(false);
	}
	
	public synchronized void stop(boolean clear)
	{
		if(clear)
			clear();
		setSpeed(0);
		m_pause = true;
		m_pauseinterval = 0;
	}
	
	public synchronized void pause()
	{
		m_pause = !m_pause;
		
		if(m_pause)
		{
			m_pauseinterval = m_interval;
			setSpeed(0);
		}
		else
		{
			setSpeed(m_pauseinterval);
		}
	}
	
	public synchronized boolean getDrawLinecount()
	{
		return m_drawlinecount;
	}
	
	public synchronized void setDrawLinecount(boolean drawlinecount)
	{
		m_drawlinecount = drawlinecount;
	}

	public synchronized void setDrawLevel(boolean drawlevel)
	{
		m_drawlevel = drawlevel;
	}
	
	public synchronized void setDrawScore(boolean drawscore)
	{
		m_drawscore = drawscore;
	}
	
	public synchronized boolean getDrawScore()
	{
		return m_drawscore;
	}
	
	public synchronized boolean getDrawLevel()
	{
		return m_drawlevel;
	}

	public synchronized void inputTranslateLeft()
	{
		if(!m_pause)
		{
			translate(m_block, -1, 0);
		}
	}

	public synchronized void inputTranslateDown()
	{
		if(!m_pause)
		{
			translate(m_block, 0, 1);
		}
	}
	
	public synchronized void inputTranslateRight()
	{
		if(!m_pause)
		{
			translate(m_block, 1, 0);
		}
	}

	public synchronized void inputRotateLeft()
	{
		if(!m_pause)
		{
			rotate(m_block, -1);
		}
	}
	
	public synchronized void inputRotateRight()
	{
		if(!m_pause)
		{
			rotate(m_block, 1);
		}
	}
	
	public synchronized void inputWarp()
	{
		if(m_pause)
			return;
			
		if(m_block.getY() < 1)
			return;

		Block bdup = m_block.duplicate();
		bdup.translate(0, 1);
		while(checkCollision(bdup)==0)
		{
			bdup.translate(0, 1);
		}
		bdup.translate(0, -1);

		m_block.setPosition(bdup.getX(), bdup.getY());
		tick();
	}

	public void redrawBuffer(Rectangle r)
	{
		int i, j;
		Graphics g = m_backbuffer.getGraphics();
		
		drawBackground(g, r.x, r.y, r.width, r.height);
		
		for(j=0;j<m_height;j++)
		{
			for(i=0;i<m_width;i++)
			{
				if(m_map[i][j] != 0)
					drawPixel(g, i, j, m_map[i][j]);
			}		
		}
		
		if(m_block != null)
		{
			drawBlock(g, m_block);
			Block bdup = m_nextblock.duplicate();
			bdup.invertValue();
			bdup.setAngle(0);
			bdup.setPosition(m_width-bdup.getMap().getWidth(), 0);
			drawBlock(g, bdup);
		}

		if(getDrawScore())
			drawString(g, "Score: " + getScore(), 5, 25, Color.white, Color.black); 
		
		if(getDrawLinecount())
			drawString(g, "Lines: " + getLinecount(), 5, 45, Color.white, Color.black);
		
		if(getDrawLevel())
			drawString(g, "Level: " + getLevel(), 5, 65, Color.white, Color.black);
				
		if(m_pause)
		{
			if(m_pauseinterval == 0)
				drawString(g, "Game Over", 5, 85, Color.red, Color.black); 
			else
				drawString(g, "Pause", 5, 85, Color.white, Color.black); 
		}
		
		/*
		drawString(g, "Speed: " + m_interval, 5, 105, Color.yellow, Color.black);
		drawString(g, "Delta: " + (m_score - m_scorelevel) + " (" + (getLevel() * m_levelstep) + ")" , 5, 125, Color.yellow, Color.black);
		*/
	}

	public void update(Graphics g)
	{
    	paint(g);
	}

	public void paint(Graphics g)
	{
		Rectangle r = g.getClip().getBounds();

		if(m_backbuffer == null)
			m_backbuffer = createImage(m_width*m_blocksize, m_height*m_blocksize);
		
		redrawBuffer(r);
		g.drawImage(m_backbuffer, 0, 0, this);
	}

	public Dimension getPreferredSize()
	{
		return new Dimension(m_width*m_blocksize, m_height*m_blocksize);
	}

	public void clear()
	{
		for(int i=0;i<m_height;i++)
			setLine(i, 0);
	}
	
	public int getScore()
	{
		return m_score;
	}

	public int getLinecount()
	{
		return m_linecount;
	}

	public int getLevel()
	{
		return m_level;
	}

	// protected
	
	protected void addLevel(int dlevel)
	{
		setLevel(getLevel() + dlevel);
	}
	
	protected void setLevel(int level)
	{
		m_level = level;
	}
	
	protected void addScore(int dscore)
	{
		setScore(getScore() + dscore);
	}
	
	protected void setScore(int score)
	{
		m_score = score;
		
		if(m_score == 0)
			m_scorelevel = 0;

		// levelwechsel ueberwachen
		int delta;
		
		delta = m_score - m_scorelevel;
		if(delta > getLevel() * m_levelstep)
		{
			addLevel(1);
			m_scorelevel = m_score;
			addSpeed(-25);
		}
	}
	
	protected void addLinecount(int dlinecount)
	{
		m_linecount += dlinecount;
	}
	
	protected void setLinecount(int linecount)
	{
		m_linecount = linecount;
	}
	
	protected void addSpeed(long dspeed)
	{
		if(m_interval + dspeed > 50)
		{
			setSpeed(m_interval + dspeed);
		}
	}
	
	protected void setSpeed(long speed)
	{
		m_interval = speed;

		if(speed == 0)
		{
			m_thread = null;
		}
		else if(m_thread == null)
		{
			m_thread = new MyThread();
			m_thread.start();
		}
	}

	protected void translate(Block block, int x, int y)
	{
		Block bdup = block.duplicate();
		
		bdup.translate(x, y);
		if(checkCollision(bdup)==0)
		{
			block.translate(x, y);
		}
	}

	protected void rotate(Block block, int dir)
	{
		Block bdup = block.duplicate();
		
		bdup.rotate(dir);
		if(checkCollision(bdup)==0)
		{
			block.rotate(dir);
		}
	}

	/*
	   -1 = Wand Kollision
		0 = Keine Kollision
		1 = Block Kollision
	*/
	protected int checkCollision(Block block)
	{
		int x = block.getX();
		int y = block.getY();
		
		if(x < 0 || y < 0)
			return -1;
			
		if(x + block.getMap().getWidth() > m_width || y + block.getMap().getHeight() > m_height)
			return -1;
		
		for(int j=0;j<block.getMap().getHeight();j++)
			for(int i=0;i<block.getMap().getWidth();i++)
			{
				if(block.getMap().get(i, j) != 0)
				{
					if(m_map[x + i][y + j] != 0)
					{
						return 1;
					}
				}				
			}
			
		return 0;
	}

	protected void setPixel(int x, int y, int value)
	{
		if(x >= 0 && y >= 0 && x < m_width && y < m_height)
		{
			m_map[x][y] = value;
		}
	}

	protected void setLine(int line, int value)
	{
		for(int i=0;i<m_width;i++)
			setPixel(i, line, value);
	}

	protected void drawString(Graphics g, String s, int x, int y, Color color, Color shadow)
	{
		Color oldcolor = g.getColor(); // push old color
		Font oldfont = g.getFont();	// push old font
		
		g.setFont(new Font("Courier", Font.PLAIN | Font.BOLD, 18));
		g.setColor(shadow);
		g.drawString(s, x+1, y+1);
		g.setColor(color);
		g.drawString(s, x, y);	
		
		g.setColor(oldcolor); // pop old color
		g.setFont(oldfont);
	}

	protected void drawBackground(Graphics g, int x, int y, int width, int height)
	{
		g.clearRect(x, y, width, height);
	}

	protected void drawPixel(Graphics g, int x, int y, int value)	// negative values werden anders gezeichnet (z.b. invertiert)
	{
		Color c;
		
		c = getColor(value);
		g.setColor(c);
		if(value < 0)
			g.fill3DRect(x*m_blocksize, y*m_blocksize, m_blocksize-1, m_blocksize-1, false);
		else
			g.fill3DRect(x*m_blocksize, y*m_blocksize, m_blocksize-1, m_blocksize-1, true);
		
		int offset = 3;
		g.fill3DRect(x*m_blocksize+offset, y*m_blocksize+offset, m_blocksize-offset*2, m_blocksize-offset*2, true);
	}

	protected void drawBlock(Graphics g, Block block)
	{
		int x, y, b;
		
		x = block.getX();
		y = block.getY();
		for(int j=0;j<block.getMap().getHeight();j++)
			for(int i=0;i<block.getMap().getWidth();i++)
			{
				b = block.getMap().get(i, j);
				if(b != 0)
					drawPixel(g, i+x, j+y, block.getMap().get(i, j));
			}
	}
	
	// kopiert den block auf die tetris karte (wird aufgerufen sobald ein block den boden beruehrt)
	protected void stampBlock(Block block)
	{
		int x, y, b;
		
		x = block.getX();
		y = block.getY();
		for(int j=0;j<block.getMap().getHeight();j++)
			for(int i=0;i<block.getMap().getWidth();i++)
			{
				b = block.getMap().get(i, j);
				if(b != 0)
					setPixel(i+x, j+y, b);
			}
	}
	
	// lscht linie und verschiebt die darberliegenden blcke nach unten
	protected void removeLine(int line)
	{
		int i;
		if(line > 0)
		{
			for(int j=line;j>0;j--)
			{
				for(i=0;i<m_width;i++)
				{
					setPixel(i, j, m_map[i][j-1]);
				}
			}
		}
		
		setLine(0, 0);
	}
	
	// prft ob eine linie komplett ist
	// wenn eine linie komplett ist werden die werte mit -1 multipliziert
	// removeLines lscht dann die markierten linien
	protected int checkLines()
	{
		int lines = 0;
		int x, i, j;

		for(j=0;j<m_height;j++)
		{
			x = 0;
			for(i=0;i<m_width;i++)
			{
				if(m_map[i][j] != 0)
				{
					x++;
				}
			}
			
			if(x >= m_width)
			{
				for(i=0;i<m_width;i++)
					m_map[i][j] *= -1;

				lines++;
			}
		}

		return lines;
	}
	
	protected void removeLines()
	{
		int j;
		
		for(j=0;j<m_height;j++)
			if(m_map[0][j] < 0)
				removeLine(j);
	}		

	protected void createNewBlock()
	{
		Random random = new Random();
		int i = random.nextInt();
		int j;
		
		if(i < 0)
			i *= -1;
		
		if(m_nextblock != null)		
			m_block = m_nextblock;
		else
			m_block = getRefBlock(i%m_blocktable.length);

		m_block.setAngle(0);
		m_block.setPosition(m_width/2-m_block.getMap().getWidth()/2, 0);
		
		i = random.nextInt();
		if(i < 0)
			i *= -1;
		
		m_nextblock = getRefBlock(i%m_blocktable.length);
		
		if(checkCollision(m_block)!=0)
		{
			stop();
			
			for(j=0;j<m_height;j++)
				for(i=0;i<m_width;i++)
					if(m_map[i][j] == 0)
						setPixel(i, j, -1);
		}
	}
	
	protected int tick()
	{
		int lines;
		Block bdup = m_block.duplicate();
		bdup.translate(0, 1);
		removeLines();
		if(checkCollision(bdup)!=0)
		{
			stampBlock(m_block);
			lines = checkLines();
			if(lines > 0)
			{
				addScore((int)Math.pow(2, lines)*m_width*getLevel());
				addLinecount(lines);
			}
			
			createNewBlock();
			return 1;
		}
		else
		{
			translate(m_block, 0, 1);
			return 0;
		}
	}
	
	protected Color getColor(int value)
	{
		Color color;

		switch(value)
		{
		case -1: // ROT
			color = new Color(105, 22, 22);
			break;
		case  1:
			color = Color.red;
			break;
		case -2: // BLAU
			color = new Color(41, 25, 106);
			break;
		case 2:
			color = new Color(0, 128, 255);
			break;
		case -3: // GRN
			color = new Color(38, 190, 38);
			break;
		case 3:
			color = Color.green;
			break;
		case -4:  // GELB
			color = new Color(210, 210, 34);
			break;
		case 4:
			color = Color.yellow;
			break;
		case -5: // ORANGE 2
			color = new Color(200, 86, 36);
			break;
		case 5:
			color = new Color(255, 78, 0);
			break;
		case -6: // ORANGE
			color = new Color(217, 150, 43);
			break;
		case 6:
			color = Color.orange;
			break;
		case -7: // CYAN
			color = new Color(32, 141, 185);
			break;
		case 7:
			color = Color.cyan;
			break; 
		case 0: // SCHWARZ
		default:
			color = Color.black;
			break;
		};
		
		return color;
	}
	
	protected Block getRefBlock(int index)
	{
		return m_blocktable[index % m_blocktable.length];
	}
	
	protected Block getBlock(int index)
	{
		return getRefBlock(index).duplicate();
	}
	
	// Var
	
	protected Block m_block = null;
	protected Block m_nextblock = null;
	protected int m_map[][];
	protected int m_width, m_height;
	protected int m_blocksize;
	protected Image m_backbuffer = null;
	protected int m_score = 0;
	protected int m_scorelevel = 0;
	protected int m_linecount = 0;
	protected int m_level = 0;
	protected final int m_levelstep = 350;
	
	protected long m_interval = 0;
	protected MyThread m_thread = null;
	
	private boolean m_pause = true;
	private long m_pauseinterval = 0;
	
	private boolean m_drawscore = true;
	private boolean m_drawlinecount = true;
	private boolean m_drawlevel = true;
	
	// private

	private class MyThread extends Thread
	{
		public synchronized void run()
		{
 			MyThread myThread = (MyThread)Thread.currentThread();
   			while(m_thread == myThread)
   			{
	        	try
	        	{
            		Thread.sleep(m_interval);
        		} catch (InterruptedException e){ }
				
				if(!m_pause)
					tick();
				repaint();
			}
		}
	}

	// initBlocks
	protected Block[] m_blocktable;
	protected void initBlocks()
	{
		int i = 0;
		m_blocktable = new Block[7];
		Blockmap[] m;

		// Block 1

		// x
		// x
		// xx
		
		m = new Blockmap[4];

		m[0] = new Blockmap(2, 3);
		m[1] = new Blockmap(3, 2);
		m[2] = new Blockmap(2, 3);
		m[3] = new Blockmap(3, 2);

		m[0].set(0, 0, 1);
		m[0].set(0, 1, 1);
		m[0].set(0, 2, 1); m[0].set(1, 2, 1);

		m[1].set(0, 0, 1); m[1].set(1, 0, 1); m[1].set(2, 0, 1);
		m[1].set(0, 1, 1);

		m[2].set(0, 0, 1); m[2].set(1, 0, 1);
					       m[2].set(1, 1, 1);
					       m[2].set(1, 2, 1);

											  m[3].set(2, 0, 1);
		m[3].set(0, 1, 1); m[3].set(1, 1, 1); m[3].set(2, 1, 1);

		m_blocktable[i++] = new Block(0,0,m);
		
		// Block 2

		// xx
		// xx
		
		m = new Blockmap[1];
		
		m[0] = new Blockmap(2, 2);
		
		m[0].set(0, 0, 2); m[0].set(1, 0, 2);
		m[0].set(0, 1, 2); m[0].set(1, 1, 2);
		
		m_blocktable[i++] = new Block(0, 0, m);
		
		// Block 3
		
		// x
		// xx
		// x
		
		m = new Blockmap[4];
		
		m[0] = new Blockmap(2, 3);
		m[1] = new Blockmap(3, 2);
		m[2] = new Blockmap(2, 3);
		m[3] = new Blockmap(3, 2);
		
		m[0].set(0, 0, 3);
		m[0].set(0, 1, 3); m[0].set(1, 1, 3);
		m[0].set(0, 2, 3);
		
		m[1].set(0, 0, 3); m[1].set(1, 0, 3); m[1].set(2, 0, 3);
						   m[1].set(1, 1, 3);
	
						   m[2].set(1, 0, 3);
		m[2].set(0, 1, 3); m[2].set(1, 1, 3);
						   m[2].set(1, 2, 3);
	
						   m[3].set(1, 0, 3);
		m[3].set(0, 1, 3); m[3].set(1, 1, 3); m[3].set(2, 1, 3);
	
		m_blocktable[i++] = new Block(0, 0, m);
		
		// Block 4
		
		// x
		// xx
		//  x
		
		m = new Blockmap[2];
		
		m[0] = new Blockmap(2, 3);
		m[1] = new Blockmap(3, 2);
		
		m[0].set(0, 0, 4);
		m[0].set(0, 1, 4); m[0].set(1, 1, 4);
						   m[0].set(1, 2, 4);

						   m[1].set(1, 0, 4); m[1].set(2, 0, 4);
		m[1].set(0, 1, 4); m[1].set(1, 1, 4);
						   
		m_blocktable[i++] = new Block(0, 0, m);
						   
		// Block 5
		
		//  x
		// xx
		// x 
		
		m = new Blockmap[2];
		
		m[0] = new Blockmap(2, 3);
		m[1] = new Blockmap(3, 2);
		
						   m[0].set(1, 0, 5);
		m[0].set(0, 1, 5); m[0].set(1, 1, 5);
		m[0].set(0, 2, 5);

		m[1].set(0, 0, 5); m[1].set(1, 0, 5);
						   m[1].set(1, 1, 5); m[1].set(2, 1, 5);
						   
		m_blocktable[i++] = new Block(0, 0, m);
		
		// Block 6
		
		// x
		// x
		// x
		// x
		
		m = new Blockmap[2];
		
		m[0] = new Blockmap(1, 4);
		m[1] = new Blockmap(4, 1);
		
		m[0].set(0, 0, 6);
		m[0].set(0, 1, 6);
		m[0].set(0, 2, 6);
		m[0].set(0, 3, 6);

		m[1].set(0, 0, 6); m[1].set(1, 0, 6);  m[1].set(2, 0, 6);  m[1].set(3, 0, 6); 
				
		m_blocktable[i++] = new Block(0, 0, m);
		
		// Block 7
		
		//  x
		//  x
		// xx
				
		m = new Blockmap[4];

		m[0] = new Blockmap(2, 3);
		m[1] = new Blockmap(3, 2);
		m[2] = new Blockmap(2, 3);
		m[3] = new Blockmap(3, 2);

		
						   m[0].set(1, 0, 7);
						   m[0].set(1, 1, 7);
		m[0].set(0, 2, 7); m[0].set(1, 2, 7);
		
		m[1].set(0, 0, 7);
		m[1].set(0, 1, 7); m[1].set(1, 1, 7); m[1].set(2, 1, 7);
		
		m[2].set(0, 0, 7); m[2].set(1, 0, 7);
		m[2].set(0, 1, 7);
		m[2].set(0, 2, 7);
		
		m[3].set(0, 0, 7); m[3].set(1, 0, 7); m[3].set(2, 0, 7);
											  m[3].set(2, 1, 7);
											  
		m_blocktable[i++] = new Block(0, 0, m);
	}
}
