import java.awt.*;
import java.awt.Graphics.*;
import java.awt.image.*;

public class Tetris2D extends Tetris
{
	Tetris2D(int width, int height, int blocksize)
	{
		m_backgroundimage = getBufferedImage("tetris.jpg", this);
		if(m_backgroundimage != null)
			m_backgroundimagerect = new Rectangle(0, 0, m_backgroundimage.getWidth(), m_backgroundimage.getHeight());
		
		init(width, height, blocksize);
	}

	protected BufferedImage m_backgroundimage = null;
	protected Rectangle m_backgroundimagerect;

	protected void drawBackground(Graphics g, int x, int y, int width, int height)
	{
		Graphics2D g2D = (Graphics2D)g;
		Paint oldpaint = g2D.getPaint();
		
		if(m_backgroundimage != null)
		{
			g2D.setPaint(new TexturePaint(m_backgroundimage, m_backgroundimagerect));
		}
		else
		{
			g2D.setPaint(new GradientPaint(0, 0, getForeground(), 0, m_height*m_blocksize, getBackground(), false));
		}

		g2D.fillRect(x, y, width, height);
		g2D.setPaint(oldpaint);
	}

	protected void drawPixel(Graphics g, int x, int y, int value)
	{
		Graphics2D g2D = (Graphics2D)g;
		float alpha;
		GradientPaint gradient = new GradientPaint(x*m_blocksize, y*m_blocksize, getColor(-value), x*m_blocksize, y*m_blocksize+m_blocksize, getColor(value), false);
		Composite oldcomposite = g2D.getComposite();
		Paint oldpaint = g2D.getPaint();
		
		if(value < 0)
			alpha = 0.5f;
		else
			alpha = 0.80f;
		
		g2D.setColor(getColor(value));
		g2D.setPaint(gradient);
		g2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
		g2D.fillRect(x*m_blocksize, y*m_blocksize, m_blocksize, m_blocksize);
		g2D.draw3DRect(x*m_blocksize, y*m_blocksize, m_blocksize-1, m_blocksize-1, true);
		
		g2D.setComposite(oldcomposite);
		g2D.setPaint(oldpaint);
	}

	protected void drawString(Graphics g, String s, int x, int y, Color color, Color shadow)
	{
		Graphics2D g2D = (Graphics2D)g;

		Color oldcolor = g2D.getColor(); // push old color
		Font oldfont = g2D.getFont();	// push old font
		Composite oldcomposite = g2D.getComposite();

		g2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.4f));
		
		g2D.setFont(new Font("Courier", Font.PLAIN | Font.BOLD, 15));
		g2D.setColor(shadow);
		g2D.drawString(s, x+2, y+2);
		g2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.6f));
		g2D.drawString(s, x+1, y+1);

		g2D.setColor(color);
		g2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1.0f));
		g2D.drawString(s, x, y);	
		
		g2D.setColor(oldcolor); // pop old color
		g2D.setFont(oldfont);
		g2D.setComposite(oldcomposite);
	}
		
	// code von: http://www.apl.jhu.edu/~hall/java/Java2D-Tutorial.html
	
	protected static BufferedImage getBufferedImage(String imageFile, Component c)
	{
		Image image = c.getToolkit().getImage(imageFile);
		waitForImage(image, c);
		
		if(image.getWidth(c) < 1 || image.getHeight(c) < 1)
			return null;
		
		BufferedImage bufferedImage = new BufferedImage(image.getWidth(c), image.getHeight(c), BufferedImage.TYPE_INT_RGB);
		Graphics2D g2d = bufferedImage.createGraphics();
		g2d.drawImage(image, 0, 0, c);
	    return(bufferedImage);
	}

	protected static boolean waitForImage(Image image, Component c)
	{
		MediaTracker tracker = new MediaTracker(c);
		tracker.addImage(image, 0);
		try
		{
		tracker.waitForAll();
		}
		catch(InterruptedException ie) {}
		return(!tracker.isErrorAny());
	}

	protected static boolean waitForImages(Image[] images, Component c)
	{
		MediaTracker tracker = new MediaTracker(c);
		
		for(int i=0; i<images.length; i++)
		tracker.addImage(images[i], 0);
		try 
		{
			tracker.waitForAll();
		} catch(InterruptedException ie) {}
		
		return(!tracker.isErrorAny());
	}
}